""" Twisted Protocol for the TC2100

To use the Protocol, subclass :py:class:`ThermometerProtocol` and implement
its ``observation_received()`` method to handle messages.
"""
from abc import abstractmethod
from typing import TextIO
from csv import DictWriter
import struct

from twisted.internet.protocol import Protocol

from tc2100.observation import Observation


class ThermometerProtocol(Protocol):
    """ Twisted Protocol for the TC2100

    Subclasses should override the
    :py:meth:`ThermometerProtocol.observation_received` method, which is
    called once for each observation received.
    """
    _header = b"\x65\x14"
    _num_padbytes = 3
    _trailer = b"\x0d\x0a"
    _msglen = 2 + 3 + 2 + Observation.size()
    _data_start = 2 + 3
    _data_len = Observation.size()

    def __init__(self):
        super().__init__()
        self._buf = b""

    @abstractmethod
    def observation_received(self, observation: Observation) -> None:
        """ Callback for received Observation

        :param observation: Temperature update
        """

    def dataReceived(self, data: bytes) -> None:
        self._buf += data

        while self._framing_detector():
            try:
                msg = Observation.from_bytes(
                    self._buf[self._data_start:self._data_start +
                              self._data_len])
                self._buf = self._buf[self._msglen:]
            except (ValueError, struct.error):
                # if we failed to decode, our framing is off
                self._buf = self._buf[1:]
                continue

            self.observation_received(msg)

    def connectionLost(self, reason=None):
        self._buf = b""

    def _framing_detector(self) -> bool:
        """ Perform message boundary detection

        Serial protocols lack inherent framing---divisions between message
        boundaries. Reads might begin in the middle of a message. This
        method synchronizes to the message boundary by advancing `octets`
        until the start of message is found.

        :param data: bytes which may contain a message, not necessarily at
               start
        """

        while len(self._buf) >= self._msglen:
            dstart = self._buf[0:len(self._header)]
            dend = self._buf[self._msglen-len(self._trailer):self._msglen]
            if dstart == self._header and dend == self._trailer:
                return True

            self._buf = self._buf[1:]

        return False


class ThermometerToCSVProtocol(ThermometerProtocol):
    """ Output thermometer measurements as CSV

    .. py:attribute:: file_handle

        An open text stream where output will be written. Files must be
        opened with the ``newline=''`` option.
    """
    def __init__(self, file_handle: TextIO):
        super().__init__()
        self._writer = DictWriter(file_handle,
                                  fieldnames=Observation.field_names())
        self._writer.writeheader()

    def observation_received(self, observation: Observation) -> None:
        self._writer.writerow(observation.as_dict())



""" Temperature observation message """

import enum
import math
from struct import Struct
from datetime import datetime
from datetime import timezone
from typing import Tuple, NamedTuple


@enum.unique
class TemperatureUnit(enum.IntEnum):
    """ A unit of temperature

    .. py:attribute:: C

       Degrees Celsius

    .. py:attribute:: F

       Degrees Fahrenheit

    .. py:attribute:: K

       Kelvin
    """

    C = 1
    F = 2
    K = 3

    def __str__(self):
        return str(self.name)


@enum.unique
class ThermocoupleType(enum.IntEnum):
    """ A type of thermocouple

    .. py:attribute:: K

       K Type

    .. py:attribute:: J

       J Type

    .. py:attribute:: T

       T Type

    .. py:attribute:: E

       E Type

    .. py:attribute:: R

       R Type

    .. py:attribute:: S

       S Type

    .. py:attribute:: N

       N Type
    """

    K = 1
    J = 2
    T = 3
    E = 4
    R = 5
    S = 6
    N = 7

    def __str__(self):
        return str(self.name)


class MeterTime(NamedTuple):
    """ Elapsed time indication from the thermometer

    The TC2100 represents time as ``hours : minutes : seconds`` from
    boot-up. The maximum representable time is ``255:59:59``. `MeterTime`
    objects are immutable.

    .. py:attribute:: hours

        Elapsed time since meter bootup, in hours

    .. py:attribute:: minutes

        Elapsed time since meter bootup, in minutes

    .. py:attribute:: seconds

        Elapsed time since meter bootup, in seconds
    """
    hours: int
    minutes: int
    seconds: int

    _format = Struct("!3B")

    def __str__(self):
        return "%03d:%02d:%02d" % self

    def to_bytes(self) -> bytes:
        """ Convert to wireline representation

        :return: Packed byte representation of this MeterTime
        """
        return self._format.pack(self.hours, self.minutes, self.seconds)

    @classmethod
    def from_bytes(cls, octets: bytes) -> 'MeterTime':
        """ Convert from wireline representation

        :param octets: Packed byte representation of this message
        :return: A decoded time. An exception is thrown if the decoding
                 fails.
        """
        return MeterTime(*cls._format.unpack(octets))

    @classmethod
    def size(cls) -> int:
        """ Get packed size of this message

        :return: Size of the packed bytes generated by
                 :py:meth:`MeterTime.to_bytes`
        """
        return cls._format.size


class Observation(NamedTuple):
    """ Temperature observation message from the thermometer

    Observations are time-tagged in Python, using the current value of the
    system's real-time clock, when they are converted :py:meth:`from_bytes()`.
    Otherwise, an Observation contains only data output by the thermometer.

    .. py:attribute:: system_time

        Time, according to the system clock, when this Observation was
        de-serialized. Always a timezone-aware `datetime` in the UTC time zone.

    .. py:attribute:: meter_time

        Time, according to the meter, that the temperature measurement was
        taken.

    .. py:attribute:: thermocouple_type

        Thermocouple type used to make this measurement.

    .. py:attribute:: unit

        The TC2100 outputs temperature in the units that it was configured to
        display. This field represents the temperature units for the
        :py:meth:`temperatures`.

    .. py:attribute:: temperature_ch1

        Temperature value on channel 1, or `math.nan` if the measurement is
        not valid.

    .. py:attribute:: temperature_ch2

        Temperature value on channel 2, or `math.nan` if the measurement is
        invalid.
    """

    system_time: datetime or None
    meter_time: MeterTime or None
    thermocouple_type: ThermocoupleType or str
    unit: TemperatureUnit or str
    temperature_ch1: float
    temperature_ch2: float

    _num_channels = 2
    _flag_valid = 0x08
    _flag_invalid = 0x40
    _flag_negative = 0x80
    _invalid_placeholder = -32768
    _mask_lowbyte_only = 0x0F
    _units_highbyte = 0x80
    _tzero = MeterTime(0, 0, 0)
    _format = Struct(
        "!"
        "2h"    # measurements, in tens of display units
        "B"     # thermocouple type (enumerated)
        "B"     # display unit (enumerated)
        "2B"    # channel validity / status flags
    )

    @property
    def temperatures(self) -> Tuple[float, float]:
        """ Obtain all temperature measurements

        :return: Temperature measurements from each channel
        """
        return self.temperature_ch1, self.temperature_ch2

    def as_dict(self) -> dict:
        """ Convert to dict

        :return: This Observation, expressed as a mutable dict.
        """
        return self._asdict()

    def to_bytes(self) -> bytes:
        """ Convert to wireline representation

        The conversion will be lossy if, for example, the channel temperatures
        are not evenly divisible by 0.1.

        :return: Packed byte representation of this Observation
        """
        out_fields = list()
        out_fields.extend([self._encode_temperature_value(v)
                           for v in self.temperatures])
        if isinstance(self.thermocouple_type, str):
            thermtype = ThermocoupleType[self.thermocouple_type]
        else:
            thermtype = self.thermocouple_type
        out_fields.append(int(thermtype))

        if isinstance(self.unit, str):
            unit = TemperatureUnit[self.unit]
        else:
            unit = self.unit
        out_fields.append(int(unit) | 0x80)

        out_fields.extend([self._encode_temperature_flag(v)
                           for v in self.temperatures])

        meter_time = self.meter_time or self._tzero
        return self._format.pack(*out_fields) + meter_time.to_bytes()

    @classmethod
    def from_bytes(cls, octets: bytes) -> 'Observation':
        """ Convert from wireline representation

        :param octets: Packed byte representation of this message
        :return: A decoded message. An exception is thrown if the decoding
                 fails.
        """
        meter_time = MeterTime.from_bytes(octets[-3:])
        decode = cls._format.unpack(octets[0:-3])
        therm_type = ThermocoupleType(decode[2] & cls._mask_lowbyte_only)
        display_unit = TemperatureUnit(decode[3] & cls._mask_lowbyte_only)
        temperatures = [cls._decode_temperature(decode[i], decode[i + 4])
                        for i in range(0, cls._num_channels)]
        return Observation(datetime.now(timezone.utc),
                           meter_time,
                           therm_type,
                           display_unit,
                           temperatures[0],
                           temperatures[1])

    @classmethod
    def size(cls) -> int:
        """ Get packed size of this message

        :return Size of the packed byte representation of an Observation,
                as generated by :py:meth:`Observation.to_bytes`
        """
        return cls._format.size + MeterTime.size()

    @classmethod
    def field_names(cls) -> Tuple:
        """ Get all field names

        :return Field names of an Observation, in an order suitable for
                writing to a CSV file
        """
        return cls._fields

    @classmethod
    def _decode_temperature(cls, value, valid_flag) -> float:
        if valid_flag & cls._flag_valid:
            val = float(value) / 10.0
            if valid_flag & cls._flag_negative:
                val = -val
            return val
        return math.nan

    @classmethod
    def _encode_temperature_value(cls, temperature) -> int:
        if not math.isfinite(temperature):
            return cls._invalid_placeholder
        return int(math.fabs(temperature) * 10)

    @classmethod
    def _encode_temperature_flag(cls, temperature) -> int:
        if not math.isfinite(temperature):
            return cls._flag_invalid

        flag = cls._flag_valid
        if temperature < 0:
            flag |= cls._flag_negative
        return flag



""" Auto-discover compatible thermometers

This module searches all serial ports on your system and auto-detects
compatible thermometers.

>>> import tc2100.discover.discover
>>> ports = tc2100.discover.discover()
"""

from typing import List, Iterable
from serial.tools.list_ports import comports


VENDOR_ID = (0x10c4, )
""" USB Vendor IDs for the TC2100 """


PRODUCT_ID = (0xea60, )
""" USB Product IDs for the TC2100
"""


def discover(ports: Iterable = None) -> List:
    """ Find all attached digital thermometers

    For this method to work, the USB vendor ID and product ID must be listed
    in `VENDOR_ID` and `PRODUCT_ID`, respectively.

    :param: ports:   List of serial ports, from pyserial. Omit to autoscan.
    :return: List of discovered thermometers. If none are found, returns an
             empty list.
    """
    if ports is None:
        ports = comports(include_links=True)

    return list(filter(_is_tc2100, ports))


def _is_tc2100(port) -> bool:
    try:
        if port.vid not in VENDOR_ID or port.pid not in PRODUCT_ID:
            return False
    except AttributeError:
        return False
    return True



"""
Dump TC2100 temperature readings to CSV
"""

import argparse
import sys
import os
from typing import TextIO

from twisted.internet import reactor
from twisted.internet.serialport import SerialPort

from tc2100.protocol import ThermometerToCSVProtocol
from tc2100.discover import discover
from tc2100.version import __version__ as version


_application_name = 'tc2100dump'


def main():
    """ Run the program """
    parser = get_arg_parser()
    args = parser.parse_args()
    if args.version:
        print(version_string(), file=sys.stderr)
        sys.exit(0)

    port_name = args.port
    if not port_name:
        # Autodiscover
        ports = discover()
        if not ports:
            print("Cannot find a TC2100 compatible digital thermometer.\n\n"
                  "If you have one attached and powered on, check that its\n"
                  "USB product ID and vendor ID are listed in the discover\n"
                  "module. You can manually specify a serial port address\n"
                  "with the --port option.", file=sys.stderr)
            sys.exit(1)
        port_name = ports[0].device

    run_dump_to_csv(port_name, args.out)


def get_arg_parser() -> argparse.ArgumentParser:
    """ Obtain the argparse for the executable

    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        description="Dump values from a digital thermometer")
    parser.add_argument("--port", type=str, required=False,
                        help="Path to serial port, like /dev/ttyUSB0 or COM1")
    parser.add_argument("--out", type=str, required=False,
                        help="Output file. Use '-' for stdout.")
    parser.add_argument("--version", action='store_true',
                        help="Print version and exit")

    return parser


def run_dump_to_csv(port: str, file_name: str = None):
    """ Connect to a TC2100 and dump its output to a file

    :param port: Serial port
    :param file_name: Destination file
    :return: Zero on success, or non-zero for error
    """

    if not file_name or file_name == '-':
        _make_stdout_linebuffered()
        _run_dump(port, sys.stdout)
        return

    with open(file_name, mode='w', newline='', buffering=1) as outfile:
        _run_dump(port, outfile)


def version_string() -> str:
    """ Obtain a version string for this executable

    :return Version string
    """
    return "%s version %s" % (_application_name, version)


def _make_stdout_linebuffered() -> None:
    try:
        sys.stdout.reconfigure(line_buffering=True)
    except TypeError:
        sys.stdout = os.fdopen(0, 'w', 1)
    except AttributeError:
        sys.stdout = os.fdopen(0, 'w', 1)


def _run_dump(port: str, file_handle: TextIO) -> None:
    SerialPort(ThermometerToCSVProtocol(file_handle=file_handle),
               port, reactor, baudrate=9600)
    reactor.run()


if __name__ == '__main__':
    # execute only if run as the entry point into the program
    main()



"""
tc2100: Receive data from a compatible USB digital thermometer

This module contains a parser for the TC2100 series of digital thermometers.
The parser is implemented as a `twisted` protocol, which enables a variety
of interesting use-cases for real-time temperature data.

This module includes :py:class:`ThermometerToCSVProtocol`, which dumps the
streaming temperature data to a CSV file. To do something else with the data,
subclass :py:class:`ThermometerProtocol` like so:

.. code-block:: python

   from twisted.internet import reactor
   from twisted.internet.serialport import SerialPort
   from tc2100 import ThermometerProtocol, discover

   class ThermPrinter(ThermometerProtocol):
     def observation_received(self, observation):
       print(observation.temperatures)

   SerialPort(ThermPrinter(), discover()[0].device, reactor, baudrate=9600)
   reactor.run()

A more detailed example can be found in :py:mod:`tc2100.__main__`.
"""

from .version import __version__        # noqa:F401
from .version import __author__         # noqa:F401
from .version import __copyright__      # noqa:F401
from .version import __license__        # noqa:F401

from .discover import discover          # noqa:F401
from .protocol import ThermometerProtocol          # noqa:F401
from .protocol import ThermometerToCSVProtocol     # noqa:F401
from .observation import Observation, MeterTime    # noqa:F401
from .observation import TemperatureUnit           # noqa:F401
from .observation import ThermocoupleType          # noqa:F401



from tc2100.discover import discover, PRODUCT_ID, VENDOR_ID

from collections import namedtuple

FakePort = namedtuple('FakePort', field_names=('vid', 'pid'))
FakeNotUsb = namedtuple('FakeNotUsb', field_names=('com',))


def test_discover():
    out = discover(list())
    assert len(out) == 0

    mockports = [FakePort(vid=VENDOR_ID[0], pid=PRODUCT_ID[0]),
                 FakePort(vid=0, pid=0),
                 FakeNotUsb(com='/dev/ttyACM0')]
    out = discover(mockports)
    assert len(out) == 1
    assert out[0].vid == VENDOR_ID[0]

    out = discover((mockports[1], ))
    assert len(out) == 0



from tc2100.__main__ import version_string, get_arg_parser, \
    _make_stdout_linebuffered


def test_version_string():
    assert isinstance(version_string(), str)


def test_arg_parser():
    get_arg_parser()


def test_make_stdout_unbuffered():
    _make_stdout_linebuffered()



import math

from tc2100.observation import Observation, ThermocoupleType, TemperatureUnit,\
    MeterTime

from . import datastrings


def test_temp_unit_strings():
    assert str(TemperatureUnit.K) == 'K'
    assert int(TemperatureUnit.K) == 3


def test_thermocouple_strings():
    assert str(ThermocoupleType.K) == 'K'
    assert int(ThermocoupleType.K) == 1


def test_metertime_create():
    m = MeterTime(255, 3, 2)
    assert m.hours == 255
    assert m.minutes == 3
    assert m.seconds == 2
    assert str(m) == "255:03:02"


def test_metertime_packunpack():
    m = MeterTime.from_bytes(datastrings.NO_DATA[13:16])
    assert m.hours == 0
    assert m.seconds == 42
    assert m.minutes == 6


def test_decode_no_data():
    msg = Observation.from_bytes(datastrings.NO_DATA[5:5+11])
    assert msg.thermocouple_type == ThermocoupleType.N
    assert msg.temperatures[0] != msg.temperatures[0]
    assert msg.temperatures[1] != msg.temperatures[1]
    assert msg.unit == TemperatureUnit.C
    assert str(msg.meter_time) == "000:06:42"
    msgdict = msg.as_dict()
    assert str(msgdict['meter_time']) == "000:06:42"


def test_decode_ch1():
    msg = Observation.from_bytes(datastrings.CH1[5:5+11])
    assert msg.thermocouple_type == ThermocoupleType.N
    assert msg.temperatures[0] == 20.8
    assert msg.temperatures[1] != msg.temperatures[1]
    assert msg.unit == TemperatureUnit.C


def test_decode_ch2():
    msg = Observation.from_bytes(datastrings.CH2[5:5+11])
    assert msg.thermocouple_type == ThermocoupleType.N
    assert msg.temperatures[0] != msg.temperatures[0]
    assert msg.temperatures[1] == 20.4
    assert msg.unit == TemperatureUnit.C


def test_decode_kelvin():
    msg = Observation.from_bytes(datastrings.BOTH_K[5:5+11])
    assert msg.temperatures[0] == datastrings.BOTH_K_CH1
    assert msg.temperatures[1] == datastrings.BOTH_K_CH2
    assert msg.unit == TemperatureUnit.K


def test_decode_negative_temperature():
    msg = Observation.from_bytes(datastrings.NEGATIVE_14P1[5:5+11])
    assert msg.temperature_ch1 == -14.1

    msg2 = Observation.from_bytes(msg.to_bytes())
    assert msg2.temperature_ch1 == -14.1


def test_decode_reencode():
    msg = Observation.from_bytes(datastrings.BOTH[5:5+11])

    assert msg.thermocouple_type == ThermocoupleType.K
    assert msg.temperatures[0] == 22.8
    assert msg.temperatures[1] == 23.2
    assert msg.unit == TemperatureUnit.C
    assert str(msg.meter_time) == "002:28:30"

    outb = msg.to_bytes()
    cmp = datastrings.BOTH[5:5+11]
    assert outb == cmp


def test_convert_synthetic():
    msg = Observation(temperature_ch1=math.nan,
                      temperature_ch2=math.nan,
                      unit='K', thermocouple_type='N',
                      meter_time=None, system_time=None)
    out = msg.to_bytes()
    inp = Observation.from_bytes(out)

    assert inp.temperatures[0] != inp.temperatures[0]
    assert inp.temperatures[1] != inp.temperatures[1]
    assert inp.thermocouple_type == ThermocoupleType.N
    assert inp.unit == TemperatureUnit.K
    assert str(inp.meter_time) == "000:00:00"



from setuptools import setup, find_packages
import os

# Single-source version information
__version__ = None
__license__ = None
__author__ = None
repo_root = os.path.join(os.path.dirname(__file__))
vers_file = os.path.join(repo_root, "tc2100", "version.py")
exec(open(vers_file).read())

with open('README.md') as readme:
    long_description = readme.read()


with open('requirements.txt') as requirements:
    dependencies = [line.strip() for line in requirements]


setup(
    name='tc2100',
    version=__version__,
    license=__license__,
    description='Receive data from a compatible USB digital thermometer',
    long_description=long_description,
    long_description_content_type='text/markdown',

    author=__author__,
    author_email='3526918+cbs228@users.noreply.github.com',
    url='https://github.com/cbs228/tc2100',

    platforms=['any'],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'Framework :: Twisted',
    ],
    keywords='thermometer tc2100 thermocouple 10c4:ea60 serial driver',

    entry_points={
        'console_scripts': ['tc2100dump=tc2100.__main__:main'],
    },

    packages=find_packages(),

    install_requires=dependencies,
)



import io

from tc2100.protocol import ThermometerProtocol, ThermometerToCSVProtocol
from tc2100.observation import Observation

from . import datastrings


class BufferingThermProto(ThermometerProtocol):
    def __init__(self):
        super().__init__()
        self.messages = []

    def observation_received(self, observation: Observation):
        self.messages.append(observation)


def test_framing_empty():
    buf = BufferingThermProto()
    buf.dataReceived(b"")

    assert len(buf.messages) == 0
    assert len(buf._buf) == 0


def test_framing_leadingzeros():
    buf = BufferingThermProto()
    buf.dataReceived(datastrings.FRAMING_LEADZERO)

    assert len(buf.messages) == 1
    assert len(buf._buf) == 0
    assert buf.messages[0].temperatures[0] == datastrings.BOTH_K_CH1
    assert buf.messages[0].temperatures[1] == datastrings.BOTH_K_CH2


def test_framing_bad_header():
    buf = BufferingThermProto()
    buf.dataReceived(datastrings.FRAMING_BAD_HEADER)

    assert len(buf.messages) == 1
    assert len(buf._buf) == 0
    assert buf.messages[0].temperatures[0] == datastrings.BOTH_K_CH1
    assert buf.messages[0].temperatures[1] == datastrings.BOTH_K_CH2


def test_framing_trail_header():
    buf = BufferingThermProto()
    buf.dataReceived(datastrings.FRAMING_TRAIL_HEADER)

    assert len(buf.messages) == 1
    assert buf._buf == b"\x65\x14\xFF"
    assert buf.messages[0].temperatures[0] == datastrings.BOTH_K_CH1
    assert buf.messages[0].temperatures[1] == datastrings.BOTH_K_CH2


def test_framing_two_messages():
    buf = BufferingThermProto()
    buf.dataReceived(datastrings.FRAMING_TWOMSGS)

    assert len(buf.messages) == 2
    assert len(buf._buf) == 0
    assert buf.messages[0].temperatures[0] == datastrings.BOTH_K_CH1
    assert buf.messages[0].temperatures[1] == datastrings.BOTH_K_CH2
    assert buf.messages[1].temperatures[0] == datastrings.BOTH_K_CH1
    assert buf.messages[1].temperatures[1] == datastrings.BOTH_K_CH2


def test_framing_bad_msg():
    buf = BufferingThermProto()
    buf.dataReceived(datastrings.BAD_UNIT)

    assert len(buf.messages) == 0


def test_csvwriter():
    buf = io.StringIO()
    conv = ThermometerToCSVProtocol(file_handle=buf)
    conv.dataReceived(datastrings.FRAMING_LEADZERO)
    conv.dataReceived(datastrings.FRAMING_LEADZERO)
    lines = buf.getvalue().splitlines()
    assert len(lines) == 3
